==> clahe-1.00.tbz: clahe-1.00/main.c <==
// astretch.c
// makes a new RGB image by  automatic histogram stretching
// heavily, and in a dirty(?) way, reuses code from resample.c 
// and clahe.c , see below,
// sorry, and thanks,
// Tue Jul 16 07:27:29 EDT 2002 <pac@next.gli.cas.cz>
// License: GPL
// Version 1.00

/*
 * ANSI C code from the article
 * "Contrast Limited Adaptive Histogram Equalization"
 * by Karel Zuiderveld, karel@cv.ruu.nl
 * in "Graphics Gems IV", Academic Press, 1994
 *
 *
 *  These functions implement Contrast Limited Adaptive Histogram Equalization.
 *  The main routine (CLAHE) expects an input image that is stored contiguously in
 *  memory;  the CLAHE output image overwrites the original input image and has the
 *  same minimum and maximum values (which must be provided by the user).
 *  This implementation assumes that the X- and Y image resolutions are an integer
 *  multiple of the X- and Y sizes of the contextual regions. A check on various other
 *  error conditions is performed.
 *
 *  #define the symbol BYTE_IMAGE to make this implementation suitable for
 *  8-bit images. The maximum number of contextual regions can be redefined
 *  by changing uiMAX_REG_X and/or uiMAX_REG_Y; the use of more than 256
 *  contextual regions is not recommended.
 *
 *  The code is ANSI-C and is also C++ compliant.
 *
 *  Author: Karel Zuiderveld, Computer Vision Research Group,
 *	     Utrecht, The Netherlands (karel@cv.ruu.nl)
 */



#include <u.h>
#include <libc.h>
#include <draw.h>
#include <memdraw.h>
#include "clahe.h"

#define ALPHA 3


void
usage(void)
{
	fprint(2, "usage:\n\tclahe [imagefile]\n");
	exits("usage");
}

static void
do_clahe(uchar **oscan, uchar **nscan, int xsize, int ysize, int channels)
{
	kz_pixel_t *pImage;
	uchar *src, *dest, old, new;
	uchar  min[3], max[3];  /* Minimums and maximums */
	int            x1, y1, x2, y2;
	long          i, x, y, off;
	int            k, status;

	switch(channels){
	case 3:
	case 4: 
		break;
	default:
		sysfatal("can't handle  %d-channel images, need 24(32)-bpp depth", channels);	
	}

	/* Get image for each channel */
pImage = malloc(xsize*ysize*sizeof(kz_pixel_t*));

	for (k = 0; k < channels; k++) {
		for (x = 0; x < xsize; x++) {
			off=channels*x+k;
			for(y=0; y<ysize; y++){  // for every point in scanline
				pImage[y*xsize+x] =  oscan[y][off];
			}
		}
				if (status=CLAHE(pImage, xsize, ysize, 0, 255, 2, 2, 256, 3.0) != 0) print("%d\n", status);

		//back to nscan
		for (x = 0; x < xsize; x++) {
			off=channels*x+k;
			for(y=0; y<ysize; y++){  // for every point in scanline
				nscan[y][off] =  pImage[y*xsize+x];
			}
		}
	} //k


} /* do_clahe */




int
max(int a, int b)
{
	if(a > b)
		return a;
	return b;
}


Memimage*
mknewimg(int xsize, int ysize, Memimage *m)
{
	int i, j, bpl, nchan;
	Memimage *new;
	uchar **oscan, **nscan;

	new = allocmemimage(Rect(0, 0, xsize, ysize), m->chan);
	if(new == nil)
		sysfatal("can't allocate new image: %r");

	oscan = malloc(Dy(m->r)*sizeof(uchar*));
	nscan = malloc(max(ysize, Dy(m->r))*sizeof(uchar*));
	if(oscan == nil || nscan == nil)
		sysfatal("can't allocate: %r");

	/* unload original image into scan lines */
	bpl = bytesperline(m->r, m->depth);
	for(i=0; i<Dy(m->r); i++){
		oscan[i] = malloc(bpl);
		if(oscan[i] == nil)
			sysfatal("can't allocate: %r");
		j = unloadmemimage(m, Rect(m->r.min.x, m->r.min.y+i, m->r.max.x, m->r.min.y+i+1), oscan[i], bpl);
		if(j != bpl)
			sysfatal("unloadmemimage");
	}

	/* allocate scan lines for destination. we do y first, so need at least Dy(m->r) lines */
	bpl = bytesperline(Rect(0, 0, xsize, Dy(m->r)), m->depth);
	for(i=0; i<max(ysize, Dy(m->r)); i++){
		nscan[i] = malloc(bpl);
		if(nscan[i] == nil)
			sysfatal("can't allocate: %r");
	}

	nchan = m->depth/8;

	do_clahe(oscan, nscan, xsize, ysize, nchan);



	/* pack data into destination */
	bpl = bytesperline(new->r, m->depth);
	for(i=0; i<ysize; i++){
		j = loadmemimage(new, Rect(0, i, xsize, i+1), nscan[i], bpl);
		if(j != bpl)
			sysfatal("loadmemimage: %r");
	}
	return new;
}


void
main(int argc, char *argv[])
{
	int i, fd, xsize, ysize, dummy;
	Rectangle rparam;
	Memimage *m, *new, *t1, *t2;
	char *file;
	ulong tchan;
	char tmp[100];
	double v;

	memimageinit();
	memset(&rparam, 0, sizeof rparam);
	xsize = ysize = 0;


	file = "<stdin>";
	fd = 0;

	switch (argc) {
	case 2:
		file = argv[1];
		fd = open(file, OREAD);
		if(fd < 0)
			sysfatal("can't open %s: %r", file);
	case 1:
		break;
	default:
		usage();
	}

	m = readmemimage(fd);
	if(m == nil)
		sysfatal("can't read %s: %r", file);

	xsize = Dx(m->r);
	ysize = Dy(m->r);

	new = nil;
	switch(m->chan){

	case RGB24:
	case RGBA32:
	case ARGB32:
	case XRGB32:
		new = mknewimg(xsize, ysize, m);
		break;

	case GREY1:
	case GREY2:
	case GREY4:
	case GREY8:
	case CMAP8:
	case RGB15:
	case RGB16:
		tchan = RGB24;
		goto Convert;

	Convert:
		/* use library to convert to byte-per-chan form, then convert back */
		t1 = allocmemimage(m->r, tchan);
		if(t1 == nil)
			sysfatal("can't allocate temporary image: %r");
		memimagedraw(t1, t1->r, m, m->r.min, nil, ZP, S);
		t2 = mknewimg(xsize, ysize, t1);
		freememimage(t1);
		new = allocmemimage(Rect(0, 0, xsize, ysize), m->chan);
		if(new == nil)
			sysfatal("can't allocate new image: %r");
		/* should do error diffusion here */
		memimagedraw(new, new->r, t2, t2->r.min, nil, ZP, S);
		freememimage(t2);
		break;

	default:
		sysfatal("can't handle channel type %s", chantostr(tmp, m->chan));
	}

	assert(new);
	if(writememimage(1, new) < 0)
		sysfatal("write error on output: %r");

	exits(nil);
}

==> clahe-1.01.tbz: clahe-1.01/main.c <==
// makes a new RGB image by  Contrast Limited Adaptive Histogram Equalization;
// heavily, and in a dirty(?) way, reuses code from resample.c 
// and clahe.c , see below,
// sorry, and thanks,
// Tue Jul 16 07:27:29 EDT 2002 <pac@next.gli.cas.cz>
// License: GPL
// Version 1.00

/*
 * ANSI C code from the article
 * "Contrast Limited Adaptive Histogram Equalization"
 * by Karel Zuiderveld, karel@cv.ruu.nl
 * in "Graphics Gems IV", Academic Press, 1994
 *
 *
 *  These functions implement Contrast Limited Adaptive Histogram Equalization.
 *  The main routine (CLAHE) expects an input image that is stored contiguously in
 *  memory;  the CLAHE output image overwrites the original input image and has the
 *  same minimum and maximum values (which must be provided by the user).
 *  This implementation assumes that the X- and Y image resolutions are an integer
 *  multiple of the X- and Y sizes of the contextual regions. A check on various other
 *  error conditions is performed.
 *
 *  #define the symbol BYTE_IMAGE to make this implementation suitable for
 *  8-bit images. The maximum number of contextual regions can be redefined
 *  by changing uiMAX_REG_X and/or uiMAX_REG_Y; the use of more than 256
 *  contextual regions is not recommended.
 *
 *  The code is ANSI-C and is also C++ compliant.
 *
 *  Author: Karel Zuiderveld, Computer Vision Research Group,
 *	     Utrecht, The Netherlands (karel@cv.ruu.nl)
 */



#include <u.h>
#include <libc.h>
#include <draw.h>
#include <memdraw.h>
#include "clahe.h"

#define ALPHA 3

double amount = 3.0;

void
usage(void)
{
	fprint(2, "usage:\n\tclahe [-a amount] [imagefile]\n");
	exits("usage");
}

static void
do_clahe(uchar **oscan, uchar **nscan, int xsize, int ysize, int channels)
{
	kz_pixel_t *pImage;
	uchar *src, *dest, old, new;
	uchar  min[3], max[3];  /* Minimums and maximums */
	int            x1, y1, x2, y2;
	long          i, x, y, off;
	int            k, status;

	switch(channels){
	case 3:
	case 4: 
		break;
	default:
		sysfatal("can't handle  %d-channel images, need 24(32)-bpp depth", channels);	
	}
/* 
	Get minimum and maximum values for each channel 

	min[0] = min[1] = min[2] = 255;
	max[0] = max[1] = max[2] = 0;

	for (x = 0; x < xsize; x++) {
		for (k = 0; k < channels; k++){
			off=channels*x+k;
			for(y=0; y<ysize; y++){  // for every point in scanline
				old =  oscan[y][off];
					if (old < min[k])
						min[k] = old;
					if (old > max[k])
						max[k] = old;
			}
		}

	} 
*/
	/* Get image for each channel */
pImage = malloc(xsize*ysize*sizeof(kz_pixel_t*));

	for (k = 0; k < channels; k++) {
		for (x = 0; x < xsize; x++) {
			off=channels*x+k;
			for(y=0; y<ysize; y++){  // for every point in scanline
				pImage[y*xsize+x] =  oscan[y][off];
			}
		}
				if (status=CLAHE(pImage, xsize, ysize, 1, 255, 2, 2, 256, amount) != 0) fprint(2,"%d\n", status);

		//back to nscan
		for (x = 0; x < xsize; x++) {
			off=channels*x+k;
			for(y=0; y<ysize; y++){  // for every point in scanline
				nscan[y][off] =  pImage[y*xsize+x];
			}
		}
	} //k


} /* do_clahe */




int
max(int a, int b)
{
	if(a > b)
		return a;
	return b;
}


Memimage*
mknewimg(int xsize, int ysize, Memimage *m)
{
	int i, j, bpl, nchan;
	Memimage *new;
	uchar **oscan, **nscan;

	new = allocmemimage(Rect(0, 0, xsize, ysize), m->chan);
	if(new == nil)
		sysfatal("can't allocate new image: %r");

	oscan = malloc(Dy(m->r)*sizeof(uchar*));
	nscan = malloc(max(ysize, Dy(m->r))*sizeof(uchar*));
	if(oscan == nil || nscan == nil)
		sysfatal("can't allocate: %r");

	/* unload original image into scan lines */
	bpl = bytesperline(m->r, m->depth);
	for(i=0; i<Dy(m->r); i++){
		oscan[i] = malloc(bpl);
		if(oscan[i] == nil)
			sysfatal("can't allocate: %r");
		j = unloadmemimage(m, Rect(m->r.min.x, m->r.min.y+i, m->r.max.x, m->r.min.y+i+1), oscan[i], bpl);
		if(j != bpl)
			sysfatal("unloadmemimage");
	}

	/* allocate scan lines for destination. we do y first, so need at least Dy(m->r) lines */
	bpl = bytesperline(Rect(0, 0, xsize, Dy(m->r)), m->depth);
	for(i=0; i<max(ysize, Dy(m->r)); i++){
		nscan[i] = malloc(bpl);
		if(nscan[i] == nil)
			sysfatal("can't allocate: %r");
	}

	nchan = m->depth/8;

	do_clahe(oscan, nscan, xsize, ysize, nchan);



	/* pack data into destination */
	bpl = bytesperline(new->r, m->depth);
	for(i=0; i<ysize; i++){
		j = loadmemimage(new, Rect(0, i, xsize, i+1), nscan[i], bpl);
		if(j != bpl)
			sysfatal("loadmemimage: %r");
	}
	return new;
}


void
main(int argc, char *argv[])
{
	int i, fd, xsize, ysize, dummy;
	Rectangle rparam;
	Memimage *m, *new, *t1, *t2;
	char *file;
	ulong tchan;
	char tmp[100];
	double v;

	ARGBEGIN{
	case 'a':
		amount=atof(ARGF());
//print("%f\n",amount);
//exits(0);
		if(amount <= 1.0)
			usage();
		break;
//	default:
//		usage();
	}ARGEND
	if(argc > 1)
		usage();
	else if(argc == 1){
		file = argv[0];
		fd = open(file, OREAD);
		if(fd < 0)
			sysfatal("can't open %s: %r", file);
	}



	memimageinit();
	memset(&rparam, 0, sizeof rparam);
	xsize = ysize = 0;

	file = "<stdin>";
	fd = 0;
	if(argc > 1)
		usage();
	else if(argc == 1){
		file = argv[0];
		fd = open(file, OREAD);
		if(fd < 0)
			sysfatal("can't open %s: %r", file);
	}


	m = readmemimage(fd);
	if(m == nil)
		sysfatal("can't read %s: %r", file);

	xsize = Dx(m->r);
	ysize = Dy(m->r);

	new = nil;
	switch(m->chan){

	case RGB24:
	case RGBA32:
	case ARGB32:
	case XRGB32:
		new = mknewimg(xsize, ysize, m);
		break;

	case GREY1:
	case GREY2:
	case GREY4:
	case GREY8:
	case CMAP8:
	case RGB15:
	case RGB16:
		tchan = RGB24;
		goto Convert;

	Convert:
		/* use library to convert to byte-per-chan form, then convert back */
		t1 = allocmemimage(m->r, tchan);
		if(t1 == nil)
			sysfatal("can't allocate temporary image: %r");
		memimagedraw(t1, t1->r, m, m->r.min, nil, ZP, S);
		t2 = mknewimg(xsize, ysize, t1);
		freememimage(t1);
		new = allocmemimage(Rect(0, 0, xsize, ysize), m->chan);
		if(new == nil)
			sysfatal("can't allocate new image: %r");
		/* should do error diffusion here */
		memimagedraw(new, new->r, t2, t2->r.min, nil, ZP, S);
		freememimage(t2);
		break;

	default:
		sysfatal("can't handle channel type %s", chantostr(tmp, m->chan));
	}

	assert(new);
	if(writememimage(1, new) < 0)
		sysfatal("write error on output: %r");

	exits(nil);
}
